//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Item/InstrumentItems.h
//! @brief     Defines class InstrumentItem and all its children
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_ITEM_INSTRUMENTITEMS_H
#define BORNAGAIN_GUI_MODEL_ITEM_INSTRUMENTITEMS_H

#include "GUI/Model/Group/SelectionDescriptor.h"
#include "GUI/Model/Group/SelectionProperty.h"
#include "GUI/Model/Item/BackgroundItems.h"
#include "GUI/Model/Item/BeamItems.h" // inheritance prevents forward declaration
#include "GUI/Model/Item/DetectorItems.h"
#include "GUI/Model/Types/AxisProperty.h"
#include "GUI/Model/Types/VectorProperty.h"
#include <functional>
#include <memory>

class BackgroundItem;
class DataItem;
class DepthProbeSimulation;
class Instrument;
class ICoordSystem;
class MaskContainerItem;
class RealDataItem;

//! Abstract base class for instrument-specific item classes.

class InstrumentItem {
public:
    InstrumentItem();
    virtual ~InstrumentItem() = default;

    //! Creates an exact copy; also ID is the same!
    InstrumentItem* createCopy() const;

    virtual void serialize(Serializer& s);

    QString id() const;
    void setId(const QString& id);

    void setInstrumentName(const QString& instrumentName);
    QString instrumentName() const;

    //! The type as how to show it on the UI. Do not use for type checking or similar!
    virtual QString instrumentType() const = 0;

    template <typename T> bool is() const { return dynamic_cast<const T*>(this) != nullptr; }

    QString description() const;
    void setDescription(const QString& description);

    virtual BeamItem* beamItem() const;
    BackgroundItem* backgroundItem() const;

    template <typename T> T* setBackgroundType();
    SelectionDescriptor<BackgroundItem*> backgroundSelection() const;

    virtual std::vector<int> shape() const = 0;

    virtual void importMasks(const MaskContainerItem*) {}

    virtual void updateToRealData(const RealDataItem* item) = 0;
    virtual bool alignedWith(const RealDataItem* item) const;

    //! The default user visible name when creating an instrument
    virtual QString defaultName() const = 0;

    virtual ICoordSystem* createCoordSystem() const = 0;

    bool withPolarizerAnalyzer() const;
    void setWithPolarizerAnalyzer(bool with);

    VECTOR_PROPERTY(polarization, Polarization);
    VECTOR_PROPERTY(analyzerDirection, AnalyzerDirection);
    DOUBLE_PROPERTY(analyzerEfficiency, AnalyzerEfficiency)
    DOUBLE_PROPERTY(analyzerTotalTransmission, AnalyzerTotalTransmission);


protected:
    explicit InstrumentItem(const QString& modelType);

    template <typename T> T* beam() const;

protected:
    QString m_id;
    QString m_name;
    QString m_description;
    bool m_withPolarizerAnalyzer;
    SelectionProperty<BackgroundItem*> m_background;
    std::unique_ptr<BeamItem> m_beam;
};


class SpecularInstrumentItem : public InstrumentItem {
public:
    SpecularInstrumentItem();

    SpecularBeamItem* beamItem() const override;

    std::vector<int> shape() const override;
    void updateToRealData(const RealDataItem* item) override;
    bool alignedWith(const RealDataItem* item) const override;
    QString defaultName() const override;
    QString instrumentType() const override;

    ICoordSystem* createCoordSystem() const override;
};


class DepthProbeInstrumentItem : public InstrumentItem {
public:
    DepthProbeInstrumentItem();
    void serialize(Serializer& s) override;

    SpecularBeamItem* beamItem() const override;

    std::vector<int> shape() const override;
    void updateToRealData(const RealDataItem* item) override;
    QString defaultName() const override;
    QString instrumentType() const override;

    // FIXME switch to base ISimulation class after InstrumentItem refactoring and
    // after ISimulation gets createCoordSystem method
    std::unique_ptr<DepthProbeSimulation> createSimulation() const;

    ICoordSystem* createCoordSystem() const override;

    AXIS_PROPERTY(zAxis);
};


class Instrument2DItem : public InstrumentItem {
public:
    void serialize(Serializer& s) override;

    DetectorItem* detector() const;
    GroupItem* detectorGroup();
    template <typename T> T* setDetectorType();
    SelectionDescriptor<DetectorItem*> detectorSelection() const;

    void importMasks(const MaskContainerItem* maskContainer) override;

    std::unique_ptr<Instrument> createInstrument() const;

    static bool isDetectorPropertyName(const QString& name);

protected:
    Instrument2DItem();

    SelectionProperty<DetectorItem*> m_detector;
};


class GISASInstrumentItem : public Instrument2DItem {
public:
    std::vector<int> shape() const override;
    void updateToRealData(const RealDataItem* item) override;
    QString defaultName() const override;
    QString instrumentType() const override;

    ICoordSystem* createCoordSystem() const override;
};


class OffSpecularInstrumentItem : public Instrument2DItem {
public:
    OffSpecularInstrumentItem();
    void serialize(Serializer& s) override;

    std::vector<int> shape() const override;
    void updateToRealData(const RealDataItem* item) override;
    QString defaultName() const override;
    QString instrumentType() const override;

    ICoordSystem* createCoordSystem() const override;

    AXIS_PROPERTY(alphaAxis);
};

//  ************************************************************************************************
//  templated functions implementation
//  ************************************************************************************************

template <typename T> T* InstrumentItem::setBackgroundType()
{
    m_background.set<T>();
    return dynamic_cast<T*>(m_background.get());
}

template <typename T> T* Instrument2DItem::setDetectorType()
{
    m_detector.set<T>();
    return dynamic_cast<T*>(m_detector.get());
}

class InstrumentItems {
public:
    ~InstrumentItems();

    template <typename T> T* addInstrument()
    {
        auto* t = new T();
        m_instruments << t;
        return t;
    }

    void removeInstrument(InstrumentItem* instrument);
    void clear();

    void serialize(Serializer& s);

    //! Inserts a deep copy (also of any non xml data in a pointwise axis)
    //! The id will not be copied, but a new unique one will be created
    //! Returns the newly created instrument.
    InstrumentItem* insertCopy(const InstrumentItem& source);

    QVector<InstrumentItem*> instrumentItems() const;
    QVector<Instrument2DItem*> instrument2DItems() const;
    QVector<InstrumentItem*>
    instrumentItems(const std::function<bool(const InstrumentItem*)>& accept) const;

    InstrumentItem* findInstrumentById(const QString& instrumentId) const;
    bool instrumentExists(const QString& instrumentId) const;

    QString suggestInstrumentName(const QString& baseName) const;
    QStringList instrumentNames() const;

private:
    QVector<InstrumentItem*> m_instruments;
};

#endif // BORNAGAIN_GUI_MODEL_ITEM_INSTRUMENTITEMS_H
