//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/DetectorAlignmentEditor.cpp
//! @brief     Implements class DetectorAlignmentEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/DetectorAlignmentEditor.h"
#include "GUI/Model/Item/RectangularDetectorItem.h"
#include "GUI/Util/ComboProperty.h"
#include "GUI/View/Edit/DoubleSpinBox.h"
#include "GUI/View/Instrument/VectorEditor.h"
#include "GUI/View/Tool/WidgetUtils.h"
#include <QComboBox>
#include <QDoubleSpinBox>
#include <QFormLayout>
#include <QGroupBox>
#include <QStackedLayout>

namespace {
QString alignmentDescription(RectangularDetector::EDetectorArrangement a)
{
    switch (a) {
    case RectangularDetector::GENERIC:
        return "Intersection of normal and detector";
    case RectangularDetector::PERPENDICULAR_TO_SAMPLE:
        return "Intersection of sample x-axis and detector";
    case RectangularDetector::PERPENDICULAR_TO_DIRECT_BEAM:
        return "Intersection of direct beam and detector";
    case RectangularDetector::PERPENDICULAR_TO_REFLECTED_BEAM:
        return "Intersection of reflected beam and detector";
    case RectangularDetector::PERPENDICULAR_TO_REFLECTED_BEAM_DPOS:
        return "Intersection of direct beam and detector";
    default:
        ASSERT(false);
    }
}

} // namespace

DetectorAlignmentEditor::DetectorAlignmentEditor(QWidget* parent, RectangularDetectorItem* item)
    : QWidget(parent)
    , m_item(item)
{
    ASSERT(m_item);
    m_formLayout = new QFormLayout(this);
    m_formLayout->setContentsMargins(0, 15, 0, 0);
    m_formLayout->setSpacing(8);

    auto* m_combo =
        GUI::Util::createSelectionCombo(this, item->detectorAlignmentSelection(), [=](int) {
            createAligmentWidgets();
            emit dataChanged();
        });
    m_formLayout->addRow("Alignment:", m_combo);

    createAligmentWidgets();
}

DoubleSpinBox* DetectorAlignmentEditor::createSpinBox(QFormLayout* parentFormLayout,
                                                      const DoubleDescriptor& d)
{
    auto* sb = GUI::Util::createSpinBox(parentFormLayout, d);
    connect(sb, qOverload<double>(&DoubleSpinBox::baseValueChanged), [=](double v) {
        if (d.get() != v) {
            d.set(v);
            emit dataChanged();
        }
    });
    return sb;
}

void DetectorAlignmentEditor::createAligmentWidgets()
{
    while (m_formLayout->rowCount() > 1)
        m_formLayout->removeRow(1);

    const QString descr = alignmentDescription(m_item->detectorAlignment());
    if (m_item->detectorAlignment() == RectangularDetector::GENERIC) {
        auto* layout = new QHBoxLayout(this);
        auto* gbox = new QGroupBox(descr, this);
        auto* form = new QFormLayout(gbox);

        createSpinBox(form, m_item->u0());
        createSpinBox(form, m_item->v0());

        auto* normalEditor = new VectorEditor(
            "Normal vector",
            "Normal of the detector plane with length equal to the sample detector distance (%1)",
            this);
        normalEditor->setVector(m_item->normalVector());
        connect(normalEditor, &VectorEditor::vectorChanged, [=]() { dataChanged(); });

        auto* directionEditor = new VectorEditor(
            "Detector axis direction vector",
            "Detector axis direction vector w.r.t. the sample coordinate system (%1)", this);
        directionEditor->setVector(m_item->directionVector());
        connect(directionEditor, &VectorEditor::vectorChanged, [=]() { dataChanged(); });

        layout->addWidget(gbox);
        layout->addWidget(normalEditor);
        layout->addWidget(directionEditor);

        m_formLayout->addRow(layout);
    } else {
        auto* gbox = new QGroupBox(descr, this);
        auto* form = new QFormLayout(gbox);
        createSpinBox(form, m_item->u0());
        createSpinBox(form, m_item->v0());
        createSpinBox(form, m_item->distance());
        m_formLayout->addRow(gbox);
    }
}
